from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterable, MutableSequence
from enum import Enum, unique, Flag
from operator import index
from sys import maxsize
from typing import Any, Final, Iterator, Literal, Self, SupportsIndex, overload

from .model import DecorationModel, DeviceModel, PartModel


__all__ = [
    "Ant",
    "BasePart",
    "BlinkingTile",
    "Bumper",
    "Cannon",
    "ConveyorBelt",
    "DashTunnel",
    "Drawbridge",
    "Fan",
    "FixedSpeedDevice",
    "Gear",
    "Goal",
    "GreenCrystal",
    "KororinCapsule",
    "Magnet",
    "MagnetSegment",
    "MagnifyingGlass",
    "MelodyTile",
    "MovingCurve",
    "MovingTile",
    "Part",
    "Press",
    "ProgressMarker",
    "Punch",
    "Scissors",
    "SeesawBlock",
    "SizeTunnel",
    "SlidingTile",
    "Spring",
    "Start",
    "Thorn",
    "TimedDevice",
    "ToyTrain",
    "TrainTrack",
    "Turntable",
    "UpsideDownBall",
    "UpsideDownStageDevice",
    "Walls",
    "Warp",
]


class BasePart(ABC):
    """Base class for all stage elements"""

    __match_args__ = ("x_pos", "y_pos", "z_pos", "x_rot", "y_rot", "z_rot")
    __slots__ = ("_x_pos", "_x_rot", "_y_pos", "_y_rot", "_z_pos", "_z_rot")

    _x_pos: float
    _x_rot: float
    _y_pos: float
    _y_rot: float
    _z_pos: float
    _z_rot: float

    def __init__(
        self,
        x_pos: float,
        y_pos: float,
        z_pos: float,
        x_rot: float,
        y_rot: float,
        z_rot: float,
    ) -> None:
        self.x_pos = x_pos
        self.x_rot = x_rot
        self.y_pos = y_pos
        self.y_rot = y_rot
        self.z_pos = z_pos
        self.z_rot = z_rot

    @property
    @abstractmethod
    def cost(self) -> int:
        """The number of kororin points that this part costs to place."""

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.x_pos!r}, {self.y_pos!r}, {self.z_pos!r}, {self.x_rot!r}, {self.y_rot!r}, {self.z_rot!r})"

    @property
    def x_pos(self) -> float:
        """Positive is right, negative is left"""
        return self._x_pos

    @x_pos.setter
    def x_pos(self, value: float, /) -> None:
        self._x_pos = value

    @property
    def x_rot(self) -> float:
        """Represented in degrees from 0 to 360
        Positive turns top face front, negative turns top face back
        """
        return self._x_rot

    @x_rot.setter
    def x_rot(self, value: float, /) -> None:
        self._x_rot = value % 360

    @property
    def y_pos(self) -> float:
        """Positive is up, negative is down"""
        return self._y_pos

    @y_pos.setter
    def y_pos(self, value: float, /) -> None:
        self._y_pos = value

    @property
    def y_rot(self) -> float:
        """Represented in degrees from 0 to 360
        Positive turns front face right, negative turns front face left
        """
        return self._y_rot

    @y_rot.setter
    def y_rot(self, value: float, /) -> None:
        self._y_rot = value % 360

    @property
    def z_pos(self) -> float:
        """Positive is front, negative is back"""
        return self._z_pos

    @z_pos.setter
    def z_pos(self, value: float, /) -> None:
        self._z_pos = value

    @property
    def z_rot(self) -> float:
        """Represented in degrees from 0 to 360
        Positive turns top face left, negative turns top face right
        """
        return self._z_rot

    @z_rot.setter
    def z_rot(self, value: float, /) -> None:
        self._z_rot = value % 360


class Part(BasePart):
    """(Usually) static model that has no behavior other than being solid."""

    __slots__ = ("_shape",)

    _shape: PartModel | DecorationModel

    def __init__(
        self,
        x_pos: float,
        y_pos: float,
        z_pos: float,
        x_rot: float,
        y_rot: float,
        z_rot: float,
        *,
        shape: PartModel | DecorationModel,
    ) -> None:
        super().__init__(x_pos, y_pos, z_pos, x_rot, y_rot, z_rot)
        self.shape = shape

    @property
    def cost(self) -> Literal[10, 15, 20]:
        if isinstance(self.shape, DecorationModel):
            return 20
        elif self.shape in frozenset(
            {
                PartModel.MagmaTile,
                PartModel.SlipperyTile,
                PartModel.StickyTile,
                PartModel.InvisibleTile,
            }
        ):
            return 15
        else:
            return 10

    @property
    def shape(self) -> PartModel | DecorationModel:
        return self._shape

    @shape.setter
    def shape(self, value: PartModel | DecorationModel, /) -> None:
        self._shape = value

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.x_pos!r}, {self.y_pos!r}, {self.z_pos!r}, {self.x_rot!r}, {self.y_rot!r}, {self.z_rot!r}, shape={self.shape!r})"


class Start(BasePart):
    __slots__ = ()

    @property
    def cost(self) -> Literal[0]:
        return 0


class Goal(BasePart):
    __slots__ = ()

    @property
    def cost(self) -> Literal[0]:
        return 0


class ProgressMarker(BasePart):
    """Either a crystal (when progress is odd) or a respawn (when progress is even)"""

    __slots__ = ("_progress",)

    _progress: Literal[1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

    def __init__(
        self,
        x_pos: float,
        y_pos: float,
        z_pos: float,
        x_rot: float,
        y_rot: float,
        z_rot: float,
        *,
        progress: Literal[1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
    ) -> None:
        super().__init__(x_pos, y_pos, z_pos, x_rot, y_rot, z_rot)
        self.progress = progress

    @property
    def cost(self) -> Literal[15, 0]:
        return 15 if self.progress & 1 else 0

    @property
    def progress(self) -> Literal[1, 2, 3, 4, 5, 6, 7, 8, 9, 10]:
        """The number displayed above the object. Controls which crystals are required to enable a respawn."""
        return self._progress

    @progress.setter
    def progress(self, value: Literal[1, 2, 3, 4, 5, 6, 7, 8, 9, 10], /):
        self._progress = value

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.x_pos!r}, {self.y_pos!r}, {self.z_pos!r}, {self.x_rot!r}, {self.y_rot!r}, {self.z_rot!r}, progress={self.progress!r})"


@unique
class Speed(Enum):
    SLOW = 55
    NORMAL = 39
    FAST = 23


@unique
class Walls(Flag):
    BACK = 8
    FRONT = 2
    LEFT = 4
    RIGHT = 1


class MovingTile(BasePart):
    """A part that moves in a straight line."""

    __slots__ = (
        "_dest_x",
        "_dest_y",
        "_dest_z",
        "_shape",
        "_speed",
        "_switch",
        "_walls",
    )

    _dest_x: float
    _dest_y: float
    _dest_z: float
    _shape: Literal[
        PartModel.Tile10x10,
        PartModel.Tile20x20,
        PartModel.TileA30x30,
        PartModel.TileA30x90,
        PartModel.Tile90x90,
        PartModel.HoleB90x90,
        PartModel.FunnelPipe,
        PartModel.StraightPipe,
    ]
    _speed: float
    _switch: bool
    _walls: Walls

    def __init__(
        self,
        x_pos: float,
        y_pos: float,
        z_pos: float,
        x_rot: float,
        y_rot: float,
        z_rot: float,
        *,
        dest_x: float,
        dest_y: float,
        dest_z: float,
        shape: Literal[
            PartModel.Tile10x10,
            PartModel.Tile20x20,
            PartModel.TileA30x30,
            PartModel.TileA30x90,
            PartModel.Tile90x90,
            PartModel.HoleB90x90,
            PartModel.FunnelPipe,
            PartModel.StraightPipe,
        ],
        speed: Speed | float = Speed.NORMAL,
        switch: bool = False,
        walls: Walls = Walls(0),
    ) -> None:
        super().__init__(x_pos, y_pos, z_pos, x_rot, y_rot, z_rot)
        self.dest_x = dest_x
        self.dest_y = dest_y
        self.dest_z = dest_z
        self._shape = shape
        self.speed = speed  # type: ignore[assignment]
        self.switch = switch
        self.walls = walls

    @property
    def cost(self) -> Literal[20, 70, 35, 40, 50, 65]:
        match self.shape:
            case PartModel.Tile10x10:
                return 20
            case PartModel.Tile20x20 | PartModel.TileA30x30:
                return 70
            case PartModel.TileA30x90:
                return 35
            case PartModel.Tile90x90 | PartModel.HoleB90x90:
                return 40
            case PartModel.FunnelPipe:
                return 50
            case PartModel.StraightPipe:
                return 65

    @property
    def dest_x(self) -> float:
        """Positive is right, negative is left"""
        return self._dest_x

    @dest_x.setter
    def dest_x(self, value: float, /) -> None:
        self._dest_x = value

    @property
    def dest_y(self) -> float:
        return self._dest_y

    @dest_y.setter
    def dest_y(self, value: float, /) -> None:
        """Positive is up, negative is down"""
        self._dest_y = value

    @property
    def dest_z(self) -> float:
        """Positive is front, negative is back"""
        return self._dest_z

    @dest_z.setter
    def dest_z(self, value: float, /) -> None:
        self._dest_z = value

    def __repr__(self) -> str:
        output: Final[list[str]] = [
            f"{type(self).__name__}({self.x_pos!r}, {self.y_pos!r}, {self.z_pos!r}, {self.x_rot!r}, {self.y_rot!r}, {self.z_rot!r}, dest_x={self.dest_x!r}, dest_y={self.dest_y!r}, dest_z={self.dest_z!r}, shape={self.shape!r}"
        ]
        if self.speed != 1.0:
            output.append(f", speed={self.speed!r}")
        if self.switch:
            output.append(f", switch={self.switch!r}")
        if self.walls:
            output.append(f", walls={self.walls!r}")
        output.append(")")
        return "".join(output)

    @property
    def shape(
        self,
    ) -> Literal[
        PartModel.Tile10x10,
        PartModel.Tile20x20,
        PartModel.TileA30x30,
        PartModel.TileA30x90,
        PartModel.Tile90x90,
        PartModel.HoleB90x90,
        PartModel.FunnelPipe,
        PartModel.StraightPipe,
    ]:
        """The appearance of this moving tile."""
        return self._shape

    @shape.setter
    def shape(
        self,
        value: Literal[
            PartModel.Tile10x10,
            PartModel.Tile20x20,
            PartModel.TileA30x30,
            PartModel.TileA30x90,
            PartModel.Tile90x90,
            PartModel.HoleB90x90,
            PartModel.FunnelPipe,
            PartModel.StraightPipe,
        ],
        /,
    ) -> None:
        if (
            value in frozenset({PartModel.FunnelPipe, PartModel.StraightPipe})
            and self.switch
        ):
            raise ValueError("Moving pipes cannot be switches")
        elif (
            value not in frozenset({PartModel.Tile20x20, PartModel.TileA30x30})
            and self.walls
        ):
            raise ValueError("Invalid shape for wall attachment")
        else:
            self._shape = value

    @property
    def speed(self) -> float:
        return self._speed

    @speed.setter
    def speed(self, value: Speed | float, /) -> None:
        match value:
            case Speed.SLOW:
                self.speed = 0.5
            case Speed.NORMAL:
                self.speed = 1.0
            case Speed.FAST:
                self.speed = 1.5
            case _:
                self._speed = value

    @property
    def switch(self) -> bool:
        """Whether this moving tile must be touched in order to move."""
        return self._switch

    @switch.setter
    def switch(self, value: bool, /) -> None:
        if value and self.shape in frozenset(
            {PartModel.FunnelPipe, PartModel.StraightPipe}
        ):
            raise ValueError("Moving pipes cannot be switches")
        else:
            self._switch = value

    @property
    def walls(self) -> Walls:
        return self._walls

    @walls.setter
    def walls(self, value: Walls, /) -> None:
        if not value or self.shape in frozenset(
            {PartModel.Tile20x20, PartModel.TileA30x30}
        ):
            self._walls = value
        else:
            raise ValueError("Invalid shape for wall attachment")


class FixedSpeedDevice(BasePart, ABC):
    """Device template for devices with speeds that can be set to Slow, Normal, or Fast"""

    __slots__ = ("_speed",)

    _speed: Speed

    def __init__(
        self,
        x_pos: float,
        y_pos: float,
        z_pos: float,
        x_rot: float,
        y_rot: float,
        z_rot: float,
        *,
        speed: Speed = Speed.NORMAL,
    ) -> None:
        super().__init__(x_pos, y_pos, z_pos, x_rot, y_rot, z_rot)
        self.speed = speed

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}({self.x_pos!r}, {self.y_pos!r}, {self.z_pos!r}, {self.x_rot!r}, {self.y_rot!r}, {self.z_rot!r}, speed={self.speed!r})"
            if self.speed != Speed.NORMAL
            else f"{type(self).__name__}({self.x_pos!r}, {self.y_pos!r}, {self.z_pos!r}, {self.x_rot!r}, {self.y_rot!r}, {self.z_rot!r})"
        )

    @property
    def speed(self) -> Speed:
        return self._speed

    @speed.setter
    def speed(self, value: Speed, /) -> None:
        self._speed = value


class MovingCurve(FixedSpeedDevice):
    __slots__ = ("_shape",)

    _shape: Literal[PartModel.CurveS, PartModel.CurveM, PartModel.CurveL]

    def __init__(
        self,
        x_pos: float,
        y_pos: float,
        z_pos: float,
        x_rot: float,
        y_rot: float,
        z_rot: float,
        *,
        shape: Literal[PartModel.CurveS, PartModel.CurveM, PartModel.CurveL],
        speed: Speed = Speed.NORMAL,
    ) -> None:
        super().__init__(x_pos, y_pos, z_pos, x_rot, y_rot, z_rot, speed=speed)
        self.shape = shape

    @property
    def cost(self) -> Literal[25]:
        return 25

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}({self.x_pos!r}, {self.y_pos!r}, {self.z_pos!r}, {self.x_rot!r}, {self.y_rot!r}, {self.z_rot!r}, shape={self.shape!r}, speed={self.speed!r})"
            if self.speed != Speed.NORMAL
            else f"{type(self).__name__}({self.x_pos!r}, {self.y_pos!r}, {self.z_pos!r}, {self.x_rot!r}, {self.y_rot!r}, {self.z_rot!r}, shape={self.shape!r})"
        )

    @property
    def shape(self) -> Literal[PartModel.CurveS, PartModel.CurveM, PartModel.CurveL]:
        return self._shape

    @shape.setter
    def shape(
        self, value: Literal[PartModel.CurveS, PartModel.CurveM, PartModel.CurveL], /
    ) -> None:
        self._shape = value


class SlidingTile(BasePart):
    __slots__ = ()

    @property
    def cost(self) -> Literal[100]:
        return 100


class ConveyorBelt(BasePart):
    __slots__ = ("_reversing",)

    _reversing: bool

    def __init__(
        self,
        x_pos: float,
        y_pos: float,
        z_pos: float,
        x_rot: float,
        y_rot: float,
        z_rot: float,
        *,
        reversing: bool = False,
    ) -> None:
        super().__init__(x_pos, y_pos, z_pos, x_rot, y_rot, z_rot)
        self.reversing = reversing

    @property
    def cost(self) -> Literal[60]:
        return 60

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}({self.x_pos!r}, {self.y_pos!r}, {self.z_pos!r}, {self.x_rot!r}, {self.y_rot!r}, {self.z_rot!r}, reversing={self.reversing!r})"
            if self.reversing
            else super().__repr__()
        )

    @property
    def reversing(self) -> bool:
        return self._reversing

    @reversing.setter
    def reversing(self, value: bool, /) -> None:
        self._reversing = value


class MagnetSegment(BasePart):
    __slots__ = ("_shape",)

    _shape: Literal[
        DeviceModel.EndMagnet,
        DeviceModel.StraightMagnet,
        DeviceModel.CurveMagnetL,
        DeviceModel.CurveMagnetS,
    ]

    def __init__(
        self,
        x_pos: float,
        y_pos: float,
        z_pos: float,
        x_rot: float,
        y_rot: float,
        z_rot: float,
        *,
        shape: Literal[
            DeviceModel.EndMagnet,
            DeviceModel.StraightMagnet,
            DeviceModel.CurveMagnetL,
            DeviceModel.CurveMagnetS,
        ],
    ) -> None:
        super().__init__(x_pos, y_pos, z_pos, x_rot, y_rot, z_rot)
        self.shape = shape

    @property
    def cost(self) -> Literal[15]:
        return 15

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.x_pos!r}, {self.y_pos!r}, {self.z_pos!r}, {self.x_rot!r}, {self.y_rot!r}, {self.z_rot!r}, shape={self.shape!r})"

    @property
    def shape(self) -> Literal[
        DeviceModel.EndMagnet,
        DeviceModel.StraightMagnet,
        DeviceModel.CurveMagnetL,
        DeviceModel.CurveMagnetS,
    ]:
        return self._shape

    @shape.setter
    def shape(
        self,
        value: Literal[
            DeviceModel.EndMagnet,
            DeviceModel.StraightMagnet,
            DeviceModel.CurveMagnetL,
            DeviceModel.CurveMagnetS,
        ],
        /,
    ) -> None:
        self._shape = value


class Magnet(BasePart, MutableSequence[MagnetSegment]):
    __slots__ = ("_segments",)

    _segments: list[MagnetSegment]

    def __init__(self, iterable: Iterable[MagnetSegment] = (), /) -> None:
        self._segments = list(iterable)

    def append(self, value: MagnetSegment) -> None:
        return self._segments.append(value)

    def clear(self) -> None:
        self._segments.clear()

    def __contains__(self, value: Any, /) -> bool:
        return value in self._segments

    @property
    def cost(self) -> int:
        return sum(segment.cost for segment in self)

    def count(self, value: Any) -> int:
        return self._segments.count(value)

    def __delitem__(self, index: SupportsIndex | slice, /) -> None:
        del self._segments[index]

    def extend(self, iterable: Iterable[MagnetSegment], /) -> None:
        self._segments.extend(iterable)

    @overload
    def __getitem__(self, index: SupportsIndex, /) -> MagnetSegment:
        pass

    @overload
    def __getitem__(self, index: slice, /) -> MutableSequence[MagnetSegment]:
        pass

    def __getitem__(
        self, index: SupportsIndex | slice, /
    ) -> MagnetSegment | MutableSequence[MagnetSegment]:
        return self._segments[index]

    def __iadd__(self, value: Iterable[MagnetSegment], /) -> Self:
        self._segments += value
        return self

    def index(
        self, value: Any, start: SupportsIndex = 0, stop: SupportsIndex = maxsize, /
    ) -> int:
        return self._segments.index(value, start, stop)

    def insert(self, index: SupportsIndex, value: MagnetSegment, /) -> None:
        return self._segments.insert(index, value)

    def __iter__(self) -> Iterator[MagnetSegment]:
        return iter(self._segments)

    def __len__(self) -> int:
        return len(self._segments)

    def pop(self, index: SupportsIndex = -1, /) -> MagnetSegment:
        return self._segments.pop(index)

    def remove(self, value: MagnetSegment, /) -> None:
        self._segments.remove(value)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._segments!r})"

    def reverse(self) -> None:
        self._segments.reverse()

    def __reversed__(self) -> Iterator[MagnetSegment]:
        return reversed(self._segments)

    @overload
    def __setitem__(self, index: SupportsIndex, value: MagnetSegment, /) -> None:
        pass

    @overload
    def __setitem__(self, slice: slice, value: Iterable[MagnetSegment], /) -> None:
        pass

    def __setitem__(
        self,
        index: SupportsIndex | slice,
        value: MagnetSegment | Iterable[MagnetSegment],
        /,
    ) -> None:
        self._segments[index] = value  # type: ignore

    @property
    def x_pos(self) -> float:
        """Positive is right, negative is left"""
        return self[-1].x_pos

    @x_pos.setter
    def x_pos(self, value: float, /) -> None:
        offset: float = value - self.x_pos
        for segment in self:
            segment.x_pos += offset

    @property
    def x_rot(self) -> float:
        """Represented in degrees from 0 to 360
        Positive turns top face front, negative turns top face back
        """
        return self[-1].x_rot

    @x_rot.setter
    def x_rot(self, value: float, /) -> None:
        """Positions of segments are not updated."""
        offset: float = value - self.x_rot
        for segment in self:
            segment.x_rot += offset

    @property
    def y_pos(self) -> float:
        """Positive is up, negative is down"""
        return self[-1].y_pos

    @y_pos.setter
    def y_pos(self, value: float, /) -> None:
        offset: float = value - self.y_pos
        for segment in self:
            segment.y_pos += offset

    @property
    def y_rot(self) -> float:
        """Represented in degrees from 0 to 360
        Positive turns front face right, negative turns front face left
        """
        return self[-1].y_rot

    @y_rot.setter
    def y_rot(self, value: float, /) -> None:
        """Positions of segments are not updated."""
        offset: float = value - self.y_rot
        for segment in self:
            segment.y_rot += offset

    @property
    def z_pos(self) -> float:
        """Positive is front, negative is back"""
        return self[-1].z_pos

    @z_pos.setter
    def z_pos(self, value: float, /) -> None:
        offset: float = value - self.z_pos
        for segment in self:
            segment.z_pos += offset

    @property
    def z_rot(self) -> float:
        """Represented in degrees from 0 to 360
        Positive turns top face left, negative turns top face right
        """
        return self[-1].z_rot

    @z_rot.setter
    def z_rot(self, value: float, /) -> None:
        """Positions of segments are not updated."""
        offset: float = value - self.z_rot
        for segment in self:
            segment.z_rot += offset


class DashTunnel(BasePart):
    __slots__ = ("_shape",)

    _shape: Literal[DeviceModel.DashTunnelA, DeviceModel.DashTunnelB]

    def __init__(
        self,
        x_pos: float,
        y_pos: float,
        z_pos: float,
        x_rot: float,
        y_rot: float,
        z_rot: float,
        *,
        shape: Literal[DeviceModel.DashTunnelA, DeviceModel.DashTunnelB],
    ) -> None:
        super().__init__(x_pos, y_pos, z_pos, x_rot, y_rot, z_rot)
        self.shape = shape

    @property
    def cost(self) -> Literal[35, 100]:
        match self.shape:
            case DeviceModel.DashTunnelA:
                return 35
            case DeviceModel.DashTunnelB:
                return 100

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.x_pos!r}, {self.y_pos!r}, {self.z_pos!r}, {self.x_rot!r}, {self.y_rot!r}, {self.z_rot!r}, shape={self.shape!r})"

    @property
    def shape(self) -> Literal[DeviceModel.DashTunnelA, DeviceModel.DashTunnelB]:
        return self._shape

    @shape.setter
    def shape(
        self, value: Literal[DeviceModel.DashTunnelA, DeviceModel.DashTunnelB], /
    ) -> None:
        self._shape = value


class SeesawBlock(BasePart):
    __slots__ = ("_auto", "_shape")

    _auto: bool
    _shape: Literal[DeviceModel.SeesawLBlock, DeviceModel.SeesawIBlock]

    def __init__(
        self,
        x_pos: float,
        y_pos: float,
        z_pos: float,
        x_rot: float,
        y_rot: float,
        z_rot: float,
        *,
        auto: bool = False,
        shape: Literal[DeviceModel.SeesawLBlock, DeviceModel.SeesawIBlock],
    ) -> None:
        super().__init__(x_pos, y_pos, z_pos, x_rot, y_rot, z_rot)
        self.auto = auto
        self.shape = shape

    @property
    def auto(self) -> bool:
        return self._auto

    @auto.setter
    def auto(self, value: bool, /) -> None:
        self._auto = value

    @property
    def cost(self) -> Literal[100]:
        return 100

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}({self.x_pos!r}, {self.y_pos!r}, {self.z_pos!r}, {self.x_rot!r}, {self.y_rot!r}, {self.z_rot!r}, auto={self.auto!r}, shape={self.shape!r})"
            if self.auto
            else f"{type(self).__name__}({self.x_pos!r}, {self.y_pos!r}, {self.z_pos!r}, {self.x_rot!r}, {self.y_rot!r}, {self.z_rot!r}, shape={self.shape!r})"
        )

    @property
    def shape(self) -> Literal[DeviceModel.SeesawLBlock, DeviceModel.SeesawIBlock]:
        return self._shape

    @shape.setter
    def shape(
        self, value: Literal[DeviceModel.SeesawLBlock, DeviceModel.SeesawIBlock], /
    ) -> None:
        self._shape = value


class Cannon(BasePart):
    __slots__ = ()

    @property
    def cost(self) -> Literal[30]:
        return 30


class Drawbridge(BasePart):
    __slots__ = ()

    @property
    def cost(self) -> Literal[50]:
        return 50


class Turntable(FixedSpeedDevice):
    __slots__ = ()

    @property
    def cost(self) -> Literal[50]:
        return 50


class Bumper(BasePart):
    __slots__ = ("_powerful",)

    _powerful: bool

    def __init__(
        self,
        x_pos: float,
        y_pos: float,
        z_pos: float,
        x_rot: float,
        y_rot: float,
        z_rot: float,
        *,
        powerful: bool = False,
    ) -> None:
        super().__init__(x_pos, y_pos, z_pos, x_rot, y_rot, z_rot)
        self.powerful = powerful

    @property
    def cost(self) -> Literal[20]:
        return 20

    @property
    def powerful(self) -> bool:
        return self._powerful

    @powerful.setter
    def powerful(self, value: bool, /) -> None:
        self._powerful = value

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}({self.x_pos!r}, {self.y_pos!r}, {self.z_pos!r}, {self.x_rot!r}, {self.y_rot!r}, {self.z_rot!r}, powerful={self.powerful!r})"
            if self.powerful
            else f"{type(self).__name__}({self.x_pos!r}, {self.y_pos!r}, {self.z_pos!r}, {self.x_rot!r}, {self.y_rot!r}, {self.z_rot!r})"
        )


class Thorn(BasePart):
    __slots__ = ()

    @property
    def cost(self) -> Literal[35]:
        return 35


class Gear(FixedSpeedDevice):
    __slots__ = ()

    @property
    def cost(self) -> Literal[50]:
        return 50


class Fan(BasePart):
    __slots__ = ("_wind_pattern",)

    _wind_pattern: Literal[
        DeviceModel.Fan, DeviceModel.PowerfulFan, DeviceModel.TimerFan
    ]

    def __init__(
        self,
        x_pos: float,
        y_pos: float,
        z_pos: float,
        x_rot: float,
        y_rot: float,
        z_rot: float,
        *,
        wind_pattern: Literal[
            DeviceModel.Fan, DeviceModel.PowerfulFan, DeviceModel.TimerFan
        ] = DeviceModel.Fan,
    ) -> None:
        super().__init__(x_pos, y_pos, z_pos, x_rot, y_rot, z_rot)
        self.wind_pattern = wind_pattern

    @property
    def cost(self) -> Literal[50]:
        return 50

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}({self.x_pos!r}, {self.y_pos!r}, {self.z_pos!r}, {self.x_rot!r}, {self.y_rot!r}, {self.z_rot!r}, wind_pattern={self.wind_pattern!r})"
            if self.wind_pattern != DeviceModel.Fan
            else f"{type(self).__name__}({self.x_pos!r}, {self.y_pos!r}, {self.z_pos!r}, {self.x_rot!r}, {self.y_rot!r}, {self.z_rot!r})"
        )

    @property
    def wind_pattern(
        self,
    ) -> Literal[DeviceModel.Fan, DeviceModel.PowerfulFan, DeviceModel.TimerFan]:
        return self._wind_pattern

    @wind_pattern.setter
    def wind_pattern(
        self,
        value: Literal[DeviceModel.Fan, DeviceModel.PowerfulFan, DeviceModel.TimerFan],
        /,
    ) -> None:
        self._wind_pattern = value


class Spring(BasePart):
    __slots__ = ()

    @property
    def cost(self) -> Literal[25]:
        return 25


@unique
class MovementTiming(Enum):
    A = 23
    B = 39
    C = 55


class TimedDevice(BasePart, ABC):
    """Device Template for devices which can have one of three timings for their movements"""

    __slots__ = ("_timing",)

    _timing: MovementTiming

    def __init__(
        self,
        x_pos: float,
        y_pos: float,
        z_pos: float,
        x_rot: float,
        y_rot: float,
        z_rot: float,
        *,
        timing: MovementTiming,
    ) -> None:
        super().__init__(x_pos, y_pos, z_pos, x_rot, y_rot, z_rot)
        self.timing = timing

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.x_pos!r}, {self.y_pos!r}, {self.z_pos!r}, {self.x_rot!r}, {self.y_rot!r}, {self.z_rot!r}, timing={self.timing!r})"

    @property
    def timing(self) -> MovementTiming:
        return self._timing

    @timing.setter
    def timing(self, value: MovementTiming, /) -> None:
        self._timing = value


class Punch(TimedDevice):
    __slots__ = ()

    @property
    def cost(self) -> Literal[30]:
        return 30


class Press(TimedDevice):
    __slots__ = ()

    @property
    def cost(self) -> Literal[50]:
        return 50


class Scissors(TimedDevice):
    __slots__ = ()

    @property
    def cost(self) -> Literal[70]:
        return 70


class MagnifyingGlass(BasePart):
    __slots__ = ()

    @property
    def cost(self) -> Literal[200]:
        return 200


class UpsideDownStageDevice(BasePart):
    __slots__ = ()

    @property
    def cost(self) -> Literal[50]:
        return 50


class UpsideDownBall(BasePart):
    __slots__ = ()

    @property
    def cost(self) -> Literal[25]:
        return 25


class SizeTunnel(BasePart):
    __slots__ = ("_size",)

    _size: Literal[DeviceModel.SmallTunnel, DeviceModel.BigTunnel]

    def __init__(
        self,
        x_pos: float,
        y_pos: float,
        z_pos: float,
        x_rot: float,
        y_rot: float,
        z_rot: float,
        *,
        size: Literal[DeviceModel.SmallTunnel, DeviceModel.BigTunnel],
    ) -> None:
        super().__init__(x_pos, y_pos, z_pos, x_rot, y_rot, z_rot)
        self.size = size

    @property
    def cost(self) -> Literal[50]:
        return 50

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.x_pos!r}, {self.y_pos!r}, {self.z_pos!r}, {self.x_rot!r}, {self.y_rot!r}, {self.z_rot!r}, size={self.size!r})"

    @property
    def size(self) -> Literal[DeviceModel.SmallTunnel, DeviceModel.BigTunnel]:
        return self._size

    @size.setter
    def size(
        self, value: Literal[DeviceModel.SmallTunnel, DeviceModel.BigTunnel], /
    ) -> None:
        self._size = value


class TrainTrack(BasePart):
    __slots__ = ("_shape",)

    _shape: Literal[
        DeviceModel.EndTracks,
        DeviceModel.LeftTracks,
        DeviceModel.RightTracks,
        DeviceModel.StraightTracks,
    ]

    def __init__(
        self,
        x_pos: float,
        y_pos: float,
        z_pos: float,
        x_rot: float,
        y_rot: float,
        z_rot: float,
        *,
        shape: Literal[
            DeviceModel.EndTracks,
            DeviceModel.LeftTracks,
            DeviceModel.RightTracks,
            DeviceModel.StraightTracks,
        ],
    ) -> None:
        super().__init__(x_pos, y_pos, z_pos, x_rot, y_rot, z_rot)
        self.shape = shape

    @property
    def cost(self) -> Literal[0, 20]:
        return 0 if self.shape is DeviceModel.EndTracks else 20

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.x_pos!r}, {self.y_pos!r}, {self.z_pos!r}, {self.x_rot!r}, {self.y_rot!r}, {self.z_rot!r}, shape={self.shape!r})"

    @property
    def shape(
        self,
    ) -> Literal[
        DeviceModel.EndTracks,
        DeviceModel.LeftTracks,
        DeviceModel.RightTracks,
        DeviceModel.StraightTracks,
    ]:
        return self._shape

    @shape.setter
    def shape(
        self,
        value: Literal[
            DeviceModel.EndTracks,
            DeviceModel.LeftTracks,
            DeviceModel.RightTracks,
            DeviceModel.StraightTracks,
        ],
        /,
    ) -> None:
        self._shape = value


class ToyTrain(BasePart, MutableSequence[TrainTrack]):
    __slots__ = ("_tracks",)

    _tracks: list[TrainTrack]

    def __init__(
        self,
        x_pos: float,
        y_pos: float,
        z_pos: float,
        x_rot: float,
        y_rot: float,
        z_rot: float,
        *,
        tracks: Iterable[TrainTrack] = (),
    ) -> None:
        super().__init__(x_pos, y_pos, z_pos, x_rot, y_rot, z_rot)
        self._tracks = list(tracks)

    def append(self, value: TrainTrack) -> None:
        return self._tracks.append(value)

    def clear(self) -> None:
        self._tracks.clear()

    def __contains__(self, value: Any, /) -> bool:
        return value in self._tracks

    @property
    def cost(self) -> int:
        return 100 + sum(track.cost for track in self)

    def count(self, value: Any) -> int:
        return self._tracks.count(value)

    def __delitem__(self, index: SupportsIndex | slice, /) -> None:
        del self._tracks[index]

    def extend(self, iterable: Iterable[TrainTrack], /) -> None:
        self._tracks.extend(iterable)

    @overload
    def __getitem__(self, index: SupportsIndex, /) -> TrainTrack:
        pass

    @overload
    def __getitem__(self, index: slice, /) -> MutableSequence[TrainTrack]:
        pass

    def __getitem__(
        self, index: SupportsIndex | slice, /
    ) -> TrainTrack | MutableSequence[TrainTrack]:
        return self._tracks[index]

    def __iadd__(self, value: Iterable[TrainTrack], /) -> Self:
        self._tracks += value
        return self

    def index(
        self, value: Any, start: SupportsIndex = 0, stop: SupportsIndex = maxsize, /
    ) -> int:
        return self._tracks.index(value, start, stop)

    def insert(self, index: SupportsIndex, value: TrainTrack, /) -> None:
        return self._tracks.insert(index, value)

    def __iter__(self) -> Iterator[TrainTrack]:
        return iter(self._tracks)

    def __len__(self) -> int:
        return len(self._tracks)

    def pop(self, index: SupportsIndex = -1, /) -> TrainTrack:
        return self._tracks.pop(index)

    def remove(self, value: TrainTrack, /) -> None:
        self._tracks.remove(value)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.x_pos!r}, {self.y_pos!r}, {self.z_pos!r}, {self.x_rot!r}, {self.y_rot!r}, {self.z_rot!r}, tracks={self._tracks!r})"

    def reverse(self) -> None:
        self._tracks.reverse()

    def __reversed__(self) -> Iterator[TrainTrack]:
        return reversed(self._tracks)

    @overload
    def __setitem__(self, index: SupportsIndex, value: TrainTrack, /) -> None:
        pass

    @overload
    def __setitem__(self, slice: slice, value: Iterable[TrainTrack], /) -> None:
        pass

    def __setitem__(
        self, index: SupportsIndex | slice, value: TrainTrack | Iterable[TrainTrack], /
    ) -> None:
        self._tracks[index] = value  # type: ignore


class Warp(BasePart):
    __slots__ = (
        "_dest_x",
        "_dest_y",
        "_dest_z",
        "_return_dest_x",
        "_return_dest_y",
        "_return_dest_z",
        "_return_x_pos",
        "_return_x_rot",
        "_return_y_pos",
        "_return_y_rot",
        "_return_z_pos",
        "_return_z_rot",
    )

    _dest_x: float
    _dest_y: float
    _dest_z: float
    _return_dest_x: float
    _return_dest_y: float
    _return_dest_z: float
    _return_x_pos: float
    _return_x_rot: float
    _return_y_pos: float
    _return_y_rot: float
    _return_z_pos: float
    _return_z_rot: float

    def __init__(
        self,
        x_pos: float,
        y_pos: float,
        z_pos: float,
        x_rot: float,
        y_rot: float,
        z_rot: float,
        *,
        dest_x: float,
        dest_y: float,
        dest_z: float,
        return_x_pos: float,
        return_y_pos: float,
        return_z_pos: float,
        return_x_rot: float,
        return_y_rot: float,
        return_z_rot: float,
        return_dest_x: float,
        return_dest_y: float,
        return_dest_z: float,
    ) -> None:
        super().__init__(x_pos, y_pos, z_pos, x_rot, y_rot, z_rot)
        self.dest_x = dest_x
        self.dest_y = dest_y
        self.dest_z = dest_z
        self.return_dest_x = return_dest_x
        self.return_dest_y = return_dest_y
        self.return_dest_z = return_dest_z
        self.return_x_pos = return_x_pos
        self.return_x_rot = return_x_rot
        self.return_y_pos = return_y_pos
        self.return_y_rot = return_y_rot
        self.return_z_pos = return_z_pos
        self.return_z_rot = return_z_rot

    @property
    def cost(self) -> Literal[25]:
        return 25

    @property
    def dest_x(self) -> float:
        """Positive is right, negative is left"""
        return self._dest_x

    @dest_x.setter
    def dest_x(self, value: float, /) -> None:
        self._dest_x = value

    @property
    def dest_y(self) -> float:
        return self._dest_y

    @dest_y.setter
    def dest_y(self, value: float, /) -> None:
        """Positive is up, negative is down"""
        self._dest_y = value

    @property
    def dest_z(self) -> float:
        """Positive is front, negative is back"""
        return self._dest_z

    @dest_z.setter
    def dest_z(self, value: float, /) -> None:
        self._dest_z = value

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.x_pos!r}, {self.y_pos!r}, {self.z_pos!r}, {self.x_rot!r}, {self.y_rot!r}, {self.z_rot!r}, dest_x={self.dest_x!r}, dest_y={self.dest_y!r}, dest_z={self.dest_z!r}, return_x_pos={self.return_x_pos}, return_y_pos={self.return_y_pos}, return_z_pos={self.return_z_pos}, return_x_rot={self.return_x_rot}, return_y_rot={self.return_y_rot}, return_z_rot={self.return_z_rot}, return_dest_x={self.return_dest_x}, return_dest_y={self.return_dest_y}, return_dest_z={self.return_dest_z})"

    @property
    def return_dest_x(self) -> float:
        """Positive is right, negative is left"""
        return self._return_dest_x

    @return_dest_x.setter
    def return_dest_x(self, value: float, /) -> None:
        self._return_dest_x = value

    @property
    def return_dest_y(self) -> float:
        return self._return_dest_y

    @return_dest_y.setter
    def return_dest_y(self, value: float, /) -> None:
        """Positive is up, negative is down"""
        self._return_dest_y = value

    @property
    def return_dest_z(self) -> float:
        """Positive is front, negative is back"""
        return self._return_dest_z

    @return_dest_z.setter
    def return_dest_z(self, value: float, /) -> None:
        self._return_dest_z = value

    @property
    def return_x_pos(self) -> float:
        """Positive is right, negative is left"""
        return self._return_x_pos

    @return_x_pos.setter
    def return_x_pos(self, value: float, /) -> None:
        self._return_x_pos = value

    @property
    def return_x_rot(self) -> float:
        """Represented in degrees from 0 to 360
        Positive turns top face front, negative turns top face back
        """
        return self._return_x_rot

    @return_x_rot.setter
    def return_x_rot(self, value: float, /) -> None:
        self._return_x_rot = value % 360

    @property
    def return_y_pos(self) -> float:
        """Positive is up, negative is down"""
        return self._return_y_pos

    @return_y_pos.setter
    def return_y_pos(self, value: float, /) -> None:
        self._return_y_pos = value

    @property
    def return_y_rot(self) -> float:
        """Represented in degrees from 0 to 360
        Positive turns front face right, negative turns front face left
        """
        return self._return_y_rot

    @return_y_rot.setter
    def return_y_rot(self, value: float, /) -> None:
        self._return_y_rot = value % 360

    @property
    def return_z_pos(self) -> float:
        """Positive is front, negative is back"""
        return self._return_z_pos

    @return_z_pos.setter
    def return_z_pos(self, value: float, /) -> None:
        self._return_z_pos = value

    @property
    def return_z_rot(self) -> float:
        """Represented in degrees from 0 to 360
        Positive turns top face left, negative turns top face right
        """
        return self._return_z_rot

    @return_z_rot.setter
    def return_z_rot(self, value: float, /) -> None:
        self._return_z_rot = value % 360


class BlinkingTile(TimedDevice):
    __slots__ = ()

    @property
    def cost(self) -> Literal[20]:
        return 20


class MelodyTile(BasePart):
    __slots__ = ("_note",)

    _note: Literal[
        DeviceModel.MelodyTileLowG,
        DeviceModel.MelodyTileLowGSharp,
        DeviceModel.MelodyTileLowA,
        DeviceModel.MelodyTileLowASharp,
        DeviceModel.MelodyTileLowB,
        DeviceModel.MelodyTileC,
        DeviceModel.MelodyTileCSharp,
        DeviceModel.MelodyTileD,
        DeviceModel.MelodyTileDSharp,
        DeviceModel.MelodyTileE,
        DeviceModel.MelodyTileF,
        DeviceModel.MelodyTileFSharp,
        DeviceModel.MelodyTileG,
        DeviceModel.MelodyTileGSharp,
        DeviceModel.MelodyTileA,
        DeviceModel.MelodyTileASharp,
        DeviceModel.MelodyTileB,
        DeviceModel.MelodyTileHighC,
        DeviceModel.MelodyTileHighCSharp,
        DeviceModel.MelodyTileHighD,
        DeviceModel.MelodyTileHighDSharp,
        DeviceModel.MelodyTileHighE,
        DeviceModel.MelodyTileHighF,
        DeviceModel.MelodyTileHighFSharp,
        DeviceModel.MelodyTileHighG,
    ]

    def __init__(
        self,
        x_pos: float,
        y_pos: float,
        z_pos: float,
        x_rot: float,
        y_rot: float,
        z_rot: float,
        *,
        note: Literal[
            DeviceModel.MelodyTileLowG,
            DeviceModel.MelodyTileLowGSharp,
            DeviceModel.MelodyTileLowA,
            DeviceModel.MelodyTileLowASharp,
            DeviceModel.MelodyTileLowB,
            DeviceModel.MelodyTileC,
            DeviceModel.MelodyTileCSharp,
            DeviceModel.MelodyTileD,
            DeviceModel.MelodyTileDSharp,
            DeviceModel.MelodyTileE,
            DeviceModel.MelodyTileF,
            DeviceModel.MelodyTileFSharp,
            DeviceModel.MelodyTileG,
            DeviceModel.MelodyTileGSharp,
            DeviceModel.MelodyTileA,
            DeviceModel.MelodyTileASharp,
            DeviceModel.MelodyTileB,
            DeviceModel.MelodyTileHighC,
            DeviceModel.MelodyTileHighCSharp,
            DeviceModel.MelodyTileHighD,
            DeviceModel.MelodyTileHighDSharp,
            DeviceModel.MelodyTileHighE,
            DeviceModel.MelodyTileHighF,
            DeviceModel.MelodyTileHighFSharp,
            DeviceModel.MelodyTileHighG,
        ],
    ) -> None:
        super().__init__(x_pos, y_pos, z_pos, x_rot, y_rot, z_rot)
        self.note = note

    @property
    def cost(self) -> Literal[20]:
        return 20

    @property
    def note(self) -> Literal[
        DeviceModel.MelodyTileLowG,
        DeviceModel.MelodyTileLowGSharp,
        DeviceModel.MelodyTileLowA,
        DeviceModel.MelodyTileLowASharp,
        DeviceModel.MelodyTileLowB,
        DeviceModel.MelodyTileC,
        DeviceModel.MelodyTileCSharp,
        DeviceModel.MelodyTileD,
        DeviceModel.MelodyTileDSharp,
        DeviceModel.MelodyTileE,
        DeviceModel.MelodyTileF,
        DeviceModel.MelodyTileFSharp,
        DeviceModel.MelodyTileG,
        DeviceModel.MelodyTileGSharp,
        DeviceModel.MelodyTileA,
        DeviceModel.MelodyTileASharp,
        DeviceModel.MelodyTileB,
        DeviceModel.MelodyTileHighC,
        DeviceModel.MelodyTileHighCSharp,
        DeviceModel.MelodyTileHighD,
        DeviceModel.MelodyTileHighDSharp,
        DeviceModel.MelodyTileHighE,
        DeviceModel.MelodyTileHighF,
        DeviceModel.MelodyTileHighFSharp,
        DeviceModel.MelodyTileHighG,
    ]:
        return self._note

    @note.setter
    def note(
        self,
        value: Literal[
            DeviceModel.MelodyTileLowG,
            DeviceModel.MelodyTileLowGSharp,
            DeviceModel.MelodyTileLowA,
            DeviceModel.MelodyTileLowASharp,
            DeviceModel.MelodyTileLowB,
            DeviceModel.MelodyTileC,
            DeviceModel.MelodyTileCSharp,
            DeviceModel.MelodyTileD,
            DeviceModel.MelodyTileDSharp,
            DeviceModel.MelodyTileE,
            DeviceModel.MelodyTileF,
            DeviceModel.MelodyTileFSharp,
            DeviceModel.MelodyTileG,
            DeviceModel.MelodyTileGSharp,
            DeviceModel.MelodyTileA,
            DeviceModel.MelodyTileASharp,
            DeviceModel.MelodyTileB,
            DeviceModel.MelodyTileHighC,
            DeviceModel.MelodyTileHighCSharp,
            DeviceModel.MelodyTileHighD,
            DeviceModel.MelodyTileHighDSharp,
            DeviceModel.MelodyTileHighE,
            DeviceModel.MelodyTileHighF,
            DeviceModel.MelodyTileHighFSharp,
            DeviceModel.MelodyTileHighG,
        ],
        /,
    ) -> None:
        self._note = value

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.x_pos!r}, {self.y_pos!r}, {self.z_pos!r}, {self.x_rot!r}, {self.y_rot!r}, {self.z_rot!r}, note={self.note!r})"


class TextBox(BasePart):
    __slots__ = ("_shape", "_text_id")

    _shape: Literal[DeviceModel.CubicTextBox, DeviceModel.WallTextBox]
    _text_id: int

    def __init__(
        self,
        x_pos: float,
        y_pos: float,
        z_pos: float,
        x_rot: float,
        y_rot: float,
        z_rot: float,
        *,
        shape: Literal[DeviceModel.CubicTextBox, DeviceModel.WallTextBox],
        text_id: SupportsIndex,
    ) -> None:
        super().__init__(x_pos, y_pos, z_pos, x_rot, y_rot, z_rot)
        self.shape = shape
        self.text_id = text_id  # type: ignore[assignment]

    @property
    def cost(self) -> Literal[0]:
        return 0

    @property
    def shape(self) -> Literal[DeviceModel.CubicTextBox, DeviceModel.WallTextBox]:
        return self._shape

    @shape.setter
    def shape(
        self, value: Literal[DeviceModel.CubicTextBox, DeviceModel.WallTextBox], /
    ) -> None:
        self._shape = value

    @property
    def text_id(self) -> int:
        return self._text_id

    @text_id.setter
    def text_id(self, value: SupportsIndex, /) -> None:
        self._text_id = index(value)


class KororinCapsule(BasePart):
    __slots__ = ()

    @property
    def cost(self) -> Literal[0]:
        return 0


class GreenCrystal(BasePart):
    __slots__ = ()

    @property
    def cost(self) -> Literal[0]:
        return 0


class Ant(BasePart):
    __slots__ = ()

    @property
    def cost(self) -> Literal[0]:
        return 0



from abc import ABC, abstractmethod

from ..stage import Stage

__all__ = ["Slot"]


class Slot(ABC):
    __slots__ = ()

    def __bool__(self) -> bool:
        """Return whether this slot is filled."""
        return self.load() is not None

    @abstractmethod
    def load(self) -> Stage | None:
        pass

    @abstractmethod
    def save(self, data: Stage | None, /) -> None:
        pass



from enum import Enum, unique
from io import BytesIO
from operator import index as ix
from os.path import basename, dirname, join
from typing import TYPE_CHECKING, Annotated, Any, Literal, SupportsIndex
from collections.abc import Mapping, Sequence

from ..stage import Stage

from . import Slot
from .xml import XmlSlot

if TYPE_CHECKING:
    from _typeshed import StrOrBytesPath
else:
    StrOrBytesPath = Any


__all__ = ["EditorPage", "get_slots", "SaveSlot"]


@unique
class EditorPage(Enum):
    ORIGINAL = 0
    FRIEND = 1
    HUDSON = 2


class SaveSlot(Slot):
    __match_args__ = ("path", "page", "index")
    __slots__ = ("_offset", "_path")

    _offset: Literal[8, 156392, 312776, 469160]
    _path: str | bytes

    def __init__(
        self,
        path: StrOrBytesPath,
        page: EditorPage,
        index: Annotated[
            SupportsIndex,
            Literal[
                1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20
            ],
        ],
    ) -> None:
        index = ix(index) - 1
        if index in range(0, 20):
            self._offset = 8 + 156864 * (index & 3)  # type: ignore[assignment]
            self._path = join(path, f"ed{(index >> 2) + 5 * page.value:02}.dat")  # type: ignore[arg-type]
        else:
            raise ValueError("index must be between 1 and 20")

    def __bool__(self) -> bool:
        try:
            with open(self._path, "rb") as f:
                f.seek(self._offset)
                return f.read(1) != b"\x00"
        except FileNotFoundError:
            return False

    def __eq__(self, other: Any, /) -> bool:
        if isinstance(other, SaveSlot):
            return self._path == other._path and self._offset == other._offset
        else:
            return NotImplemented

    def __hash__(self) -> int:
        return hash((self._offset, self._path))

    @property
    def index(
        self,
    ) -> Literal[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20]:
        return (int(basename(self._path)[2:4]) % 5 >> 2 | self._offset // 156864) + 1  # type: ignore[return-value]

    def load(self) -> Stage | None:
        try:
            with open(self._path, "rb") as f:
                f.seek(self._offset)
                with BytesIO() as b:
                    block: bytearray = bytearray()
                    while True:
                        block.clear()
                        f.readinto1(block)
                        if len(b.getbuffer()) + len(block) > 156864:
                            del block[156864 - len(b.getbuffer()) :]
                        if block[-1]:
                            b.write(block)
                        else:
                            while block:
                                if block[len(block) >> 1]:
                                    b.write(block[: (len(block) >> 1) + 1])
                                    del block[: (len(block) >> 1) + 1]
                                else:
                                    del block[len(block) >> 1 :]
                            data: bytes = b.getvalue()
                            if data:
                                return XmlSlot.deserialize(data)
                            else:
                                return None
        except FileNotFoundError:
            return None

    @property
    def page(self) -> EditorPage:
        return EditorPage(int(basename(self._path)[2:4]) // 5)

    @property
    def path(self) -> StrOrBytesPath:
        return dirname(self._path)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.path!r}, {self.page!r}, {self.index!r})"

    def save(self, data: Stage | None) -> None:
        binary: bytes = b"" if data is None else XmlSlot.serialize(data)
        if len(binary) > 156864:
            raise ValueError("serialized stage data is too large to save")
        try:
            with open(self._path, "xb") as f:
                f.write(bytes(638976))
            if data is None:
                return
        except FileExistsError:
            pass
        with open(self._path, "r+b") as f:
            f.seek(self._offset)
            f.write(binary)
            f.write(bytes(156864 - len(binary)))


def get_slots(save: StrOrBytesPath, /) -> Mapping[EditorPage, Sequence[SaveSlot]]:
    return {
        page: tuple(SaveSlot(save, page, i) for i in range(1, 21))
        for page in EditorPage
    }



from collections.abc import Iterable
from enum import Enum, unique

from .part import BasePart

__all__ = ["EditUser", "Stage", "Theme"]


@unique
class EditUser(Enum):
    BEGINNER = 0
    INTERMEDIATE = 1
    EXPERT = 2
    PROTECTED = 3
    """A stage with this edit user can only be opened in expert, and will open a blank stage when doing so."""


@unique
class Theme(Enum):
    HAUNTED_HOUSE_DARKNESS = 12
    NIGHT_CITY = 10
    THE_EMPTY_LOT = 0
    NEIGHBORS_HOUSE = 1
    SIZZLIN_DESERT = 2
    CHILL_MOUNTAIN = 3
    OCEAN_TREASURE = 4
    SPACE_STATION = 5
    STUMP_TEMPLE = 6
    TUTORIAL = 11
    CANDY_ISLAND = 7
    HAUNTED_HOUSE = 8
    CITY = 9

    @property
    def decorations_available(self) -> int:
        return {
            Theme.HAUNTED_HOUSE_DARKNESS: 8,
            Theme.NIGHT_CITY: 7,
            Theme.THE_EMPTY_LOT: 10,
            Theme.NEIGHBORS_HOUSE: 5,
            Theme.SIZZLIN_DESERT: 8,
            Theme.CHILL_MOUNTAIN: 9,
            Theme.OCEAN_TREASURE: 9,
            Theme.SPACE_STATION: 7,
            Theme.STUMP_TEMPLE: 6,
            Theme.TUTORIAL: 12,
            Theme.CANDY_ISLAND: 4,
            Theme.HAUNTED_HOUSE: 8,
            Theme.CITY: 7,
        }[self]


class Stage(set[BasePart]):
    __slots__ = ("_edit_user", "_theme", "_tilt_lock")

    _edit_user: EditUser
    _theme: Theme
    _tilt_lock: bool

    def __init__(
        self,
        iterable: Iterable[BasePart] = (),
        /,
        *,
        edit_user: EditUser = EditUser.EXPERT,
        theme: Theme = Theme.THE_EMPTY_LOT,
        tilt_lock: bool = False,
    ) -> None:
        super().__init__(iterable)
        self.edit_user = edit_user
        self.theme = theme
        self.tilt_lock = tilt_lock

    @property
    def edit_user(self) -> EditUser:
        return self._edit_user

    @edit_user.setter
    def edit_user(self, value: EditUser, /) -> None:
        self._edit_user = value

    def __repr__(self) -> str:
        return f"{type(self).__name__}({set(self)!r}, edit_user={self.edit_user!r}, theme={self.theme!r}, tilt_lock={self.tilt_lock!r})"

    @property
    def theme(self) -> Theme:
        return self._theme

    @theme.setter
    def theme(self, value: Theme, /) -> None:
        self._theme = value

    @property
    def tilt_lock(self) -> bool:
        return self._tilt_lock

    @tilt_lock.setter
    def tilt_lock(self, value: bool, /) -> None:
        self._tilt_lock = value



from collections.abc import Iterator, Sequence
from io import StringIO
from itertools import chain
from os import SEEK_END
from typing import Final
from xml.etree.ElementTree import Element, fromstring

from ..stage import EditUser, Stage, Theme
from ..stage.model import DecorationModel, DeviceModel, PartModel
from ..stage.part import (
    Ant,
    BasePart,
    BlinkingTile,
    Bumper,
    TextBox,
    Cannon,
    ConveyorBelt,
    DashTunnel,
    Drawbridge,
    Fan,
    FixedSpeedDevice,
    Gear,
    Goal,
    GreenCrystal,
    KororinCapsule,
    Magnet,
    MagnetSegment,
    MagnifyingGlass,
    MelodyTile,
    MovementTiming,
    MovingCurve,
    MovingTile,
    Part,
    Press,
    ProgressMarker,
    Punch,
    Scissors,
    SeesawBlock,
    SizeTunnel,
    SlidingTile,
    Speed,
    Spring,
    Start,
    Thorn,
    TimedDevice,
    ToyTrain,
    TrainTrack,
    Turntable,
    UpsideDownBall,
    UpsideDownStageDevice,
    Walls,
    Warp,
)
from .file import FileSlot

__all__ = ["XmlSlot"]


class XmlSlot(FileSlot):
    __slots__ = ()

    @staticmethod
    def deserialize(data: bytes) -> Stage:
        """Behavior is undefined when passed invalid stage data"""

        def get_values(element: Element, /, tag: str) -> Sequence[str]:
            return element.find(tag).text.strip().split()  # type: ignore[union-attr]

        def get_pos_rot(element: Element, /) -> Iterator[float]:
            return chain(
                map(float, get_values(element, "pos")),
                map(float, get_values(element, "rot")),
            )

        root: Final[Element] = fromstring(
            data.decode("shift_jis", "xmlcharrefreplace").replace(
                '<?xml version="1.0" encoding="SHIFT_JIS"?>',
                '<?xml version="1.0"?>\n<body>',
            )
            + "</body>"
        )
        editinfo: Final[Element] = root.find("EDITINFO")  # type: ignore[assignment]
        output: Final[Stage] = Stage(
            (),
            edit_user=(
                EditUser.PROTECTED
                if editinfo.find("EDITUSER") is None
                else EditUser(int(editinfo.find("EDITUSER").text.strip()))  # type: ignore[union-attr]
            ),
            theme=Theme(int(editinfo.find("THEME").text.strip())),  # type: ignore[union-attr]
            tilt_lock=False if editinfo.find("LOCK") is None else bool(int(editinfo.find("LOCK").text.strip())),  # type: ignore[union-attr]
        )
        groups: Final[dict[int, dict[int, Element]]] = {}
        for elem in root.find("STAGEDATA") or ():
            match elem.tag:
                case "EDIT_LIGHT" | "EDIT_BG_NORMAL":
                    continue
                case "EDIT_MAP_NORMAL":
                    output.add(
                        Part(
                            *get_pos_rot(elem),
                            shape=PartModel(int(get_values(elem, "model")[1])),
                        )
                    )
                case "EDIT_MAP_EXT":
                    output.add(
                        Part(
                            *get_pos_rot(elem),
                            shape=DecorationModel(int(get_values(elem, "model")[1])),
                        )
                    )
                case "EDIT_GIM_START":
                    output.add(Start(*get_pos_rot(elem)))
                case "EDIT_GIM_GOAL":
                    output.add(Goal(*get_pos_rot(elem)))
                case "EDIT_GIM_NORMAL":
                    match DeviceModel(int(get_values(elem, "model")[1])):
                        case DeviceModel.Crystal:
                            output.add(
                                ProgressMarker(
                                    *get_pos_rot(elem),
                                    progress=int(get_values(elem, "hook")[0]) * 2 + 1,  # type: ignore[arg-type]
                                )
                            )
                        case DeviceModel.Respawn:
                            output.add(
                                ProgressMarker(
                                    *get_pos_rot(elem),
                                    progress=int(get_values(elem, "hook")[0]) * 2 + 2,  # type: ignore[arg-type]
                                )
                            )
                        case DeviceModel.MovingTile10x10:
                            output.add(
                                MovingTile(
                                    *get_pos_rot(elem),
                                    **dict(
                                        zip(
                                            ("dest_x", "dest_y", "dest_z"),
                                            map(float, get_values(elem, "anmmov1")),
                                        )
                                    ),  # type: ignore[arg-type]
                                    shape=PartModel.Tile10x10,
                                    speed=float(get_values(elem, "anmspd")[0]),
                                )
                            )
                        case DeviceModel.MovingTile20x20:
                            output.add(
                                MovingTile(
                                    *get_pos_rot(elem),
                                    **dict(
                                        zip(
                                            ("dest_x", "dest_y", "dest_z"),
                                            map(float, get_values(elem, "anmmov1")),
                                        )
                                    ),  # type: ignore[arg-type]
                                    shape=PartModel.Tile20x20,
                                    speed=float(get_values(elem, "anmspd")[0]),
                                    walls=(
                                        Walls(0)
                                        if elem.find("hook") is None
                                        else Walls(int(get_values(elem, "hook")[1]))
                                    ),
                                )
                            )
                        case DeviceModel.MovingTile30x30:
                            output.add(
                                MovingTile(
                                    *get_pos_rot(elem),
                                    **dict(
                                        zip(
                                            ("dest_x", "dest_y", "dest_z"),
                                            map(float, get_values(elem, "anmmov1")),
                                        )
                                    ),  # type: ignore[arg-type]
                                    shape=PartModel.TileA30x30,
                                    speed=float(get_values(elem, "anmspd")[0]),
                                    walls=(
                                        Walls(0)
                                        if elem.find("hook") is None
                                        else Walls(int(get_values(elem, "hook")[1]))
                                    ),
                                )
                            )
                        case DeviceModel.MovingTile30x90:
                            output.add(
                                MovingTile(
                                    *get_pos_rot(elem),
                                    **dict(
                                        zip(
                                            ("dest_x", "dest_y", "dest_z"),
                                            map(float, get_values(elem, "anmmov1")),
                                        )
                                    ),  # type: ignore[arg-type]
                                    shape=PartModel.TileA30x90,
                                    speed=float(get_values(elem, "anmspd")[0]),
                                )
                            )
                        case DeviceModel.MovingTile90x90A:
                            output.add(
                                MovingTile(
                                    *get_pos_rot(elem),
                                    **dict(
                                        zip(
                                            ("dest_x", "dest_y", "dest_z"),
                                            map(float, get_values(elem, "anmmov1")),
                                        )
                                    ),  # type: ignore[arg-type]
                                    shape=PartModel.Tile90x90,
                                    speed=float(get_values(elem, "anmspd")[0]),
                                )
                            )
                        case DeviceModel.MovingTile90x90B:
                            output.add(
                                MovingTile(
                                    *get_pos_rot(elem),
                                    **dict(
                                        zip(
                                            ("dest_x", "dest_y", "dest_z"),
                                            map(float, get_values(elem, "anmmov1")),
                                        )
                                    ),  # type: ignore[arg-type]
                                    shape=PartModel.HoleB90x90,
                                    speed=float(get_values(elem, "anmspd")[0]),
                                )
                            )
                        case DeviceModel.MovingTile10x10Switch:
                            output.add(
                                MovingTile(
                                    *get_pos_rot(elem),
                                    **dict(
                                        zip(
                                            ("dest_x", "dest_y", "dest_z"),
                                            map(float, get_values(elem, "anmmov1")),
                                        )
                                    ),  # type: ignore[arg-type]
                                    shape=PartModel.Tile10x10,
                                    speed=float(get_values(elem, "anmspd")[0]),
                                    switch=True,
                                )
                            )
                        case DeviceModel.MovingTile20x20Switch:
                            output.add(
                                MovingTile(
                                    *get_pos_rot(elem),
                                    **dict(
                                        zip(
                                            ("dest_x", "dest_y", "dest_z"),
                                            map(float, get_values(elem, "anmmov1")),
                                        )
                                    ),  # type: ignore[arg-type]
                                    shape=PartModel.Tile20x20,
                                    speed=float(get_values(elem, "anmspd")[0]),
                                    switch=True,
                                    walls=(
                                        Walls(0)
                                        if elem.find("hook") is None
                                        else Walls(int(get_values(elem, "hook")[1]))
                                    ),
                                )
                            )
                        case DeviceModel.MovingTile30x30Switch:
                            output.add(
                                MovingTile(
                                    *get_pos_rot(elem),
                                    **dict(
                                        zip(
                                            ("dest_x", "dest_y", "dest_z"),
                                            map(float, get_values(elem, "anmmov1")),
                                        )
                                    ),  # type: ignore[arg-type]
                                    shape=PartModel.TileA30x30,
                                    speed=float(get_values(elem, "anmspd")[0]),
                                    switch=True,
                                    walls=(
                                        Walls(0)
                                        if elem.find("hook") is None
                                        else Walls(int(get_values(elem, "hook")[1]))
                                    ),
                                )
                            )
                        case DeviceModel.MovingTile30x90Switch:
                            output.add(
                                MovingTile(
                                    *get_pos_rot(elem),
                                    **dict(
                                        zip(
                                            ("dest_x", "dest_y", "dest_z"),
                                            map(float, get_values(elem, "anmmov1")),
                                        )
                                    ),  # type: ignore[arg-type]
                                    shape=PartModel.TileA30x90,
                                    speed=float(get_values(elem, "anmspd")[0]),
                                    switch=True,
                                )
                            )
                        case DeviceModel.MovingTile90x90ASwitch:
                            output.add(
                                MovingTile(
                                    *get_pos_rot(elem),
                                    **dict(
                                        zip(
                                            ("dest_x", "dest_y", "dest_z"),
                                            map(float, get_values(elem, "anmmov1")),
                                        )
                                    ),  # type: ignore[arg-type]
                                    shape=PartModel.Tile90x90,
                                    speed=float(get_values(elem, "anmspd")[0]),
                                    switch=True,
                                )
                            )
                        case DeviceModel.MovingTile90x90BSwitch:
                            output.add(
                                MovingTile(
                                    *get_pos_rot(elem),
                                    **dict(
                                        zip(
                                            ("dest_x", "dest_y", "dest_z"),
                                            map(float, get_values(elem, "anmmov1")),
                                        )
                                    ),  # type: ignore[arg-type]
                                    shape=PartModel.HoleB90x90,
                                    speed=float(get_values(elem, "anmspd")[0]),
                                    switch=True,
                                )
                            )
                        case DeviceModel.MovingFunnelPipe:
                            output.add(
                                MovingTile(
                                    *get_pos_rot(elem),
                                    **dict(
                                        zip(
                                            ("dest_x", "dest_y", "dest_z"),
                                            map(float, get_values(elem, "anmmov1")),
                                        )
                                    ),  # type: ignore[arg-type]
                                    shape=PartModel.FunnelPipe,
                                    speed=float(get_values(elem, "anmspd")[0]),
                                )
                            )
                        case DeviceModel.MovingStraightPipe:
                            output.add(
                                MovingTile(
                                    *get_pos_rot(elem),
                                    **dict(
                                        zip(
                                            ("dest_x", "dest_y", "dest_z"),
                                            map(float, get_values(elem, "anmmov1")),
                                        )
                                    ),  # type: ignore[arg-type]
                                    shape=PartModel.StraightPipe,
                                    speed=float(get_values(elem, "anmspd")[0]),
                                )
                            )
                        case DeviceModel.MovingCurveS:
                            output.add(
                                MovingCurve(
                                    *get_pos_rot(elem),
                                    shape=PartModel.CurveS,
                                    speed=Speed(int(get_values(elem, "sts")[0])),
                                )
                            )
                        case DeviceModel.MovingCurveM:
                            output.add(
                                MovingCurve(
                                    *get_pos_rot(elem),
                                    shape=PartModel.CurveM,
                                    speed=Speed(int(get_values(elem, "sts")[0])),
                                )
                            )
                        case DeviceModel.MovingCurveL:
                            output.add(
                                MovingCurve(
                                    *get_pos_rot(elem),
                                    shape=PartModel.CurveL,
                                    speed=Speed(int(get_values(elem, "sts")[0])),
                                )
                            )
                        case DeviceModel.SlidingTile:
                            output.add(SlidingTile(*get_pos_rot(elem)))
                        case DeviceModel.ConveyorBelt:
                            output.add(
                                ConveyorBelt(
                                    *get_pos_rot(elem),
                                    reversing=get_values(elem, "sts")[0] == "39",
                                )
                            )
                        case DeviceModel.DashTunnelA:
                            output.add(
                                DashTunnel(
                                    *get_pos_rot(elem),
                                    shape=DeviceModel.DashTunnelA,
                                )
                            )
                        case DeviceModel.DashTunnelB:
                            output.add(
                                DashTunnel(
                                    *get_pos_rot(elem),
                                    shape=DeviceModel.DashTunnelB,
                                )
                            )
                        case DeviceModel.SeesawLBlock:
                            output.add(
                                SeesawBlock(
                                    *get_pos_rot(elem),
                                    shape=DeviceModel.SeesawLBlock,
                                )
                            )
                        case DeviceModel.SeesawIBlock:
                            output.add(
                                SeesawBlock(
                                    *get_pos_rot(elem),
                                    shape=DeviceModel.SeesawIBlock,
                                )
                            )
                        case DeviceModel.AutoSeesawLBlock:
                            output.add(
                                SeesawBlock(
                                    *get_pos_rot(elem),
                                    auto=True,
                                    shape=DeviceModel.SeesawLBlock,
                                )
                            )
                        case DeviceModel.AutoSeesawIBlock:
                            output.add(
                                SeesawBlock(
                                    *get_pos_rot(elem),
                                    auto=True,
                                    shape=DeviceModel.SeesawIBlock,
                                )
                            )
                        case DeviceModel.Cannon:
                            output.add(Cannon(*get_pos_rot(elem)))
                        case DeviceModel.Drawbridge:
                            output.add(Drawbridge(*get_pos_rot(elem)))
                        case DeviceModel.Turntable:
                            output.add(
                                Turntable(
                                    *get_pos_rot(elem),
                                    speed=Speed(int(get_values(elem, "sts")[0])),
                                )
                            )
                        case DeviceModel.Bumper:
                            output.add(Bumper(*get_pos_rot(elem)))
                        case DeviceModel.PowerfulBumper:
                            output.add(Bumper(*get_pos_rot(elem), powerful=True))
                        case DeviceModel.Thorn:
                            output.add(Thorn(*get_pos_rot(elem)))
                        case DeviceModel.Gear:
                            output.add(
                                Gear(
                                    *get_pos_rot(elem),
                                    speed=Speed(int(get_values(elem, "sts")[0])),
                                )
                            )
                        case DeviceModel.Fan:
                            output.add(Fan(*get_pos_rot(elem)))
                        case DeviceModel.PowerfulFan:
                            output.add(
                                Fan(
                                    *get_pos_rot(elem),
                                    wind_pattern=DeviceModel.PowerfulFan,
                                )
                            )
                        case DeviceModel.TimerFan:
                            output.add(
                                Fan(
                                    *get_pos_rot(elem),
                                    wind_pattern=DeviceModel.TimerFan,
                                )
                            )
                        case DeviceModel.Spring:
                            output.add(Spring(*get_pos_rot(elem)))
                        case DeviceModel.Punch:
                            output.add(
                                Punch(
                                    *get_pos_rot(elem),
                                    timing=MovementTiming(
                                        int(get_values(elem, "sts")[0])
                                    ),
                                )
                            )
                        case DeviceModel.Press:
                            output.add(
                                Press(
                                    *get_pos_rot(elem),
                                    timing=MovementTiming(
                                        int(get_values(elem, "sts")[0])
                                    ),
                                )
                            )
                        case DeviceModel.Scissors:
                            output.add(
                                Scissors(
                                    *get_pos_rot(elem),
                                    timing=MovementTiming(
                                        int(get_values(elem, "sts")[0])
                                    ),
                                )
                            )
                        case DeviceModel.MagnifyingGlass:
                            output.add(MagnifyingGlass(*get_pos_rot(elem)))
                        case DeviceModel.UpsideDownStageDevice:
                            output.add(UpsideDownStageDevice(*get_pos_rot(elem)))
                        case DeviceModel.UpsideDownBall:
                            output.add(UpsideDownBall(*get_pos_rot(elem)))
                        case DeviceModel.SmallTunnel:
                            output.add(
                                SizeTunnel(
                                    *get_pos_rot(elem), size=DeviceModel.SmallTunnel
                                )
                            )
                        case DeviceModel.BigTunnel:
                            output.add(
                                SizeTunnel(
                                    *get_pos_rot(elem), size=DeviceModel.BigTunnel
                                )
                            )
                        case DeviceModel.BlinkingTile:
                            output.add(
                                BlinkingTile(
                                    *get_pos_rot(elem),
                                    timing=MovementTiming(
                                        int(get_values(elem, "sts")[0])
                                    ),
                                )
                            )
                        case DeviceModel.CubicTextBox:
                            output.add(
                                TextBox(
                                    *get_pos_rot(elem),
                                    shape=DeviceModel.CubicTextBox,
                                    text_id=int(get_values(elem, "sts")[0]),
                                )
                            )
                        case DeviceModel.WallTextBox:
                            output.add(
                                TextBox(
                                    *get_pos_rot(elem),
                                    shape=DeviceModel.WallTextBox,
                                    text_id=int(get_values(elem, "sts")[0]),
                                )
                            )
                        case DeviceModel.KororinCapsule:
                            output.add(KororinCapsule(*get_pos_rot(elem)))
                        case DeviceModel.GreenCrystal:
                            output.add(GreenCrystal(*get_pos_rot(elem)))
                        case DeviceModel.Ant:
                            output.add(Ant(*get_pos_rot(elem)))
                        case model if model.name.startswith("MelodyTile"):
                            output.add(MelodyTile(*get_pos_rot(elem), note=model))  # type: ignore[arg-type]
                        case _:
                            groups.setdefault(int(get_values(elem, "group")[0]), {})[
                                int(get_values(elem, "group")[1])
                            ] = elem
        for group in groups.values():
            match DeviceModel(int(get_values(group[0], "model")[1])):
                case DeviceModel.EndMagnet:
                    m: Magnet = Magnet()
                    for _, elem in sorted(group.items()):
                        m.append(
                            MagnetSegment(
                                *get_pos_rot(elem),
                                shape=DeviceModel(int(get_values(elem, "model")[1])),  # type: ignore[arg-type]
                            )
                        )
                    output.add(m)
                case DeviceModel.ToyTrain:
                    t: ToyTrain = ToyTrain(*get_pos_rot(group[0]))
                    for i, elem in sorted(group.items()):
                        if i:
                            t.append(
                                TrainTrack(
                                    *get_pos_rot(elem),
                                    shape=DeviceModel(int(get_values(elem, "model")[1])),  # type: ignore[arg-type]
                                )
                            )
                    output.add(t)
                case DeviceModel.Warp:
                    output.add(
                        Warp(
                            *get_pos_rot(group[0]),
                            **dict(
                                zip(
                                    ("dest_x", "dest_y", "dest_z"),
                                    map(float, get_values(group[0], "anmmov0")),
                                )
                            ),
                            **dict(
                                zip(
                                    (
                                        "return_x_pos",
                                        "return_y_pos",
                                        "return_z_pos",
                                        "return_x_rot",
                                        "return_y_rot",
                                        "return_z_rot",
                                    ),
                                    get_pos_rot(group[1]),
                                    strict=True,
                                )
                            ),
                            **dict(
                                zip(
                                    ("return_dest_x", "return_dest_y", "return_dest_z"),
                                    map(float, get_values(group[1], "anmmov0")),
                                )
                            ),
                        )  # type: ignore[misc]
                    )
        return output

    @staticmethod
    def serialize(stage: Stage, /) -> bytes:
        def minify(value: float, /) -> str:
            """Removes the decimal point from floats representing integers."""
            return str(int(value) if value.is_integer() else value)

        def serialize_numbers(*values: float) -> str:
            """Does not include leading or trailing spaces."""
            return " ".join(minify(value) for value in values)

        def anmtype(device: BasePart, /) -> DeviceModel:
            if isinstance(device, ProgressMarker):
                return (
                    DeviceModel.Crystal if device.progress % 2 else DeviceModel.Respawn
                )
            elif isinstance(device, MovingTile):
                match device.shape:
                    case PartModel.Tile10x10:
                        return (
                            DeviceModel.MovingTile10x10Switch
                            if device.switch
                            else DeviceModel.MovingTile10x10
                        )
                    case PartModel.Tile20x20:
                        return (
                            DeviceModel.MovingTile20x20Switch
                            if device.switch
                            else DeviceModel.MovingTile20x20
                        )
                    case PartModel.TileA30x30:
                        return (
                            DeviceModel.MovingTile30x30Switch
                            if device.switch
                            else DeviceModel.MovingTile30x30
                        )
                    case PartModel.TileA30x90:
                        return (
                            DeviceModel.MovingTile30x90Switch
                            if device.switch
                            else DeviceModel.MovingTile30x90
                        )
                    case PartModel.Tile90x90:
                        return (
                            DeviceModel.MovingTile90x90ASwitch
                            if device.switch
                            else DeviceModel.MovingTile90x90A
                        )
                    case PartModel.HoleB90x90:
                        return (
                            DeviceModel.MovingTile90x90BSwitch
                            if device.switch
                            else DeviceModel.MovingTile90x90B
                        )
                    case PartModel.FunnelPipe:
                        return DeviceModel.MovingFunnelPipe
                    case PartModel.StraightPipe:
                        return DeviceModel.MovingStraightPipe
            elif isinstance(device, MovingCurve):
                match device.shape:
                    case PartModel.CurveS:
                        return DeviceModel.MovingCurveS
                    case PartModel.CurveM:
                        return DeviceModel.MovingCurveM
                    case PartModel.CurveL:
                        return DeviceModel.MovingCurveL
            elif isinstance(device, SlidingTile):
                return DeviceModel.SlidingTile
            elif isinstance(device, ConveyorBelt):
                return DeviceModel.ConveyorBelt
            elif isinstance(device, (DashTunnel, TextBox)):
                return device.shape
            elif isinstance(device, SeesawBlock):
                if device.auto:
                    match device.shape:
                        case DeviceModel.SeesawLBlock:
                            return DeviceModel.AutoSeesawLBlock
                        case DeviceModel.SeesawIBlock:
                            return DeviceModel.AutoSeesawIBlock
                else:
                    return device.shape
            elif isinstance(device, Cannon):
                return DeviceModel.Cannon
            elif isinstance(device, Drawbridge):
                return DeviceModel.Drawbridge
            elif isinstance(device, Turntable):
                return DeviceModel.Turntable
            elif isinstance(device, Bumper):
                return (
                    DeviceModel.PowerfulBumper
                    if device.powerful
                    else DeviceModel.Bumper
                )
            elif isinstance(device, Thorn):
                return DeviceModel.Thorn
            elif isinstance(device, Gear):
                return DeviceModel.Gear
            elif isinstance(device, Fan):
                return device.wind_pattern
            elif isinstance(device, Spring):
                return DeviceModel.Spring
            elif isinstance(device, Punch):
                return DeviceModel.Punch
            elif isinstance(device, Press):
                return DeviceModel.Press
            elif isinstance(device, Scissors):
                return DeviceModel.Scissors
            elif isinstance(device, MagnifyingGlass):
                return DeviceModel.MagnifyingGlass
            elif isinstance(device, UpsideDownStageDevice):
                return DeviceModel.UpsideDownStageDevice
            elif isinstance(device, UpsideDownBall):
                return DeviceModel.UpsideDownBall
            elif isinstance(device, SizeTunnel):
                return device.size
            elif isinstance(device, BlinkingTile):
                return DeviceModel.BlinkingTile
            elif isinstance(device, MelodyTile):
                return device.note
            elif isinstance(device, KororinCapsule):
                return DeviceModel.KororinCapsule
            elif isinstance(device, GreenCrystal):
                return DeviceModel.GreenCrystal
            elif isinstance(device, Ant):
                return DeviceModel.Ant
            else:
                raise ValueError(f"part {device!r} does not have a known anmtype")

        def device_data(device: BasePart, /) -> str:
            if isinstance(device, ProgressMarker):
                return f"<hook> {(device._progress - 1) // 2} 0 </hook>\n"
            elif isinstance(device, MovingTile):
                anmmov: Final[str] = (
                    f"<anmspd> {minify(device.speed)} 0 </anmspd>\n<anmmov0> {serialize_numbers(device.x_pos, device.y_pos, device.z_pos)} </anmmov0>\n<anmmov1> {serialize_numbers(device.dest_x, device.dest_y, device.dest_z)} </anmmov1>"
                )
                if device.walls:
                    match device.shape:
                        case PartModel.Tile20x20:
                            return f"<hook> {DeviceModel.MovingTile20x20Wall.value} {device.walls.value} </hook>\n{anmmov}\n"
                        case PartModel.TileA30x30:
                            return f"<hook> {DeviceModel.MovingTile30x30Wall.value} {device.walls.value} </hook>\n{anmmov}\n"
                return anmmov
            else:
                return ""

        def sts(part: BasePart, /) -> int:
            if isinstance(part, FixedSpeedDevice):
                return part.speed.value
            elif isinstance(part, ConveyorBelt):
                return 39 if part.reversing else 23
            elif isinstance(part, TimedDevice):
                return part.timing.value
            else:
                return 7

        with StringIO(
            f'<?xml version="1.0" encoding="SHIFT_JIS"?>\n<EDITINFO>\n<THEME> {stage.theme.value} </THEME>\n<LOCK> {int(stage.tilt_lock)} </LOCK>\n<EDITUSER> {stage.edit_user.value} </EDITUSER>\n</EDITINFO>\n<STAGEDATA>\n<EDIT_BG_NORMAL>\n<model> "EBB_{stage.theme.value:02}.bin 0 </model>\n</EDIT_BG_NORMAL>\n'
        ) as output:
            output.seek(0, SEEK_END)
            group: int = 1
            for part in stage:
                if isinstance(part, Magnet):
                    for i, segment in enumerate(part):
                        output.write(
                            f'<EDIT_GIM_NORMAL>\n<model> "EGB_{stage.theme.value:02}.bin" {segment.shape.value} </model>\n<pos> {serialize_numbers(segment.x_pos, segment.y_pos, segment.z_pos)} </pos>\n<rot> {serialize_numbers(segment.x_rot, segment.y_rot, segment.z_rot)} </rot>\n<sts> 7 </sts>\n<group> {group} {i} </group>\n</EDIT_GIM_NORMAL>\n'
                        )
                    group += 1
                elif isinstance(part, ToyTrain):
                    output.write(
                        f'<EDIT_GIM_NORMAL>\n<model> "EGB_{stage.theme.value:02}.bin" {DeviceModel.ToyTrain.value} </model>\n<pos> {serialize_numbers(part.x_pos, part.y_pos, part.z_pos)} </pos>\n<rot> {serialize_numbers(part.x_rot, part.y_rot, part.z_rot)} </rot>\n<sts> 7 </sts>\n<group> {group} 0 </group>\n</EDIT_GIM_NORMAL>\n'
                    )
                    for i, track in enumerate(part, 1):
                        output.write(
                            f'<EDIT_GIM_NORMAL>\n<model> "EGB_{stage.theme.value:02}.bin" {track.shape.value} </model>\n<pos> {serialize_numbers(track.x_pos, track.y_pos, track.z_pos)} </pos>\n<rot> {serialize_numbers(track.x_rot, track.y_rot, track.z_rot)} </rot>\n<sts> 7 </sts>\n<group> {group} {i} </group>\n</EDIT_GIM_NORMAL>\n'
                        )
                    group += 1
                elif isinstance(part, Warp):
                    output.write(
                        f'<EDIT_GIM_NORMAL>\n<model> "EGB_{stage.theme.value:02}.bin" {DeviceModel.Warp.value} </model>\n<pos> {serialize_numbers(part.x_pos, part.y_pos, part.z_pos)} </pos>\n<rot> {serialize_numbers(part.x_rot, part.y_rot, part.z_rot)} </rot>\n<sts> 7 </sts>\n<anmmov0> {serialize_numbers(part.dest_x, part.dest_y, part.dest_z)} </anmmov0>\n<group> {group} 0 </group>\n</EDIT_GIM_NORMAL>\n<EDIT_GIM_NORMAL>\n<model> "EGB_{stage.theme.value:02}.bin" {DeviceModel.Warp.value} </model>\n<pos> {serialize_numbers(part.return_x_pos, part.return_y_pos, part.return_z_pos)} </pos>\n<rot> {serialize_numbers(part.return_x_rot, part.return_y_rot, part.return_z_rot)} </rot>\n<sts> 7 </sts>\n<anmmov0> {serialize_numbers(part.return_dest_x, part.return_dest_y, part.return_dest_z)} </anmmov0>\n<group> {group} 1 </group>\n</EDIT_GIM_NORMAL>\n'
                    )
                    group += 1
                else:
                    if isinstance(part, Start):
                        output.write("<EDIT_GIM_START>\n")
                    elif isinstance(part, Goal):
                        output.write("<EDIT_GIM_GOAL>\n")
                    elif isinstance(part, Part):
                        if isinstance(part.shape, DecorationModel):
                            output.write("<EDIT_MAP_EXT>\n")
                        else:
                            output.write("<EDIT_MAP_NORMAL>\n")
                    else:
                        output.write("<EDIT_GIM_NORMAL>\n")
                    if isinstance(part, Part):
                        if isinstance(part.shape, DecorationModel):
                            output.write(
                                f'<model> "EME_{stage.theme.value:02}.bin" {part.shape.value} </model>\n'
                            )
                        else:
                            output.write(
                                f'<model> "EMB_{stage.theme.value:02}.bin" {part.shape.value} </model>\n'
                            )
                    elif isinstance(part, Start):
                        output.write(
                            f'<model> "EGB_{stage.theme.value:02}.bin" 0 </model>\n'
                        )
                    elif isinstance(part, Goal):
                        output.write(
                            f'<model> "EGB_{stage.theme.value:02}.bin" 1 </model>\n'
                        )
                    else:
                        output.write(
                            f'<model> "EGB_{stage.theme.value:02}.bin" {anmtype(part).value} </model>\n'
                        )
                    output.write(
                        f"<pos> {serialize_numbers(part.x_pos, part.y_pos, part.z_pos)} </pos>\n<rot> {serialize_numbers(part.x_rot, part.y_rot, part.z_rot)} </rot>\n<sts> {sts(part)} </sts>\n"
                    )
                    try:
                        output.write(f"<anmtype> {anmtype(part).value} </anmtype>\n")
                    except ValueError:
                        pass
                    output.write(device_data(part))
                    if isinstance(part, Start):
                        output.write("</EDIT_GIM_START>\n")
                    elif isinstance(part, Goal):
                        output.write("</EDIT_GIM_GOAL>\n")
                    elif isinstance(part, Part):
                        if isinstance(part.shape, DecorationModel):
                            output.write("</EDIT_MAP_EXT>\n")
                        else:
                            output.write("</EDIT_MAP_NORMAL>\n")
                    else:
                        output.write("</EDIT_GIM_NORMAL>\n")
            output.write("</STAGEDATA>")
            return output.getvalue().encode("shift_jis", "xmlcharrefreplace")



from abc import ABC, abstractmethod
from os import remove
from os.path import isfile
from typing import TYPE_CHECKING, Any

from ..stage import Stage
from . import Slot

if TYPE_CHECKING:
    from _typeshed import StrOrBytesPath
else:
    StrOrBytesPath = Any


__all__ = ["FileSlot"]


class FileSlot(Slot, ABC):
    __match_args__ = ("path",)
    __slots__ = ("_path",)

    _path: StrOrBytesPath

    def __init__(self, path: StrOrBytesPath, /) -> None:
        self._path = path

    def __bool__(self) -> bool:
        return isfile(self.path)

    @staticmethod
    @abstractmethod
    def deserialize(data: bytes, /) -> Stage:
        pass

    def __eq__(self, other: object, /) -> bool:
        if isinstance(other, FileSlot) and (
            isinstance(other, type(self)) or isinstance(self, type(other))
        ):
            return self.path == other.path
        else:
            return NotImplemented

    def __hash__(self) -> int:
        return hash(self.path)

    def load(self) -> Stage | None:
        try:
            with open(self.path, "rb") as f:
                return self.deserialize(f.read())
        except FileNotFoundError:
            return None

    @property
    def path(self) -> StrOrBytesPath:
        return self._path

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.path!r})"

    def save(self, data: Stage | None) -> None:
        if data is None:
            remove(self.path)
        else:
            with open(self.path, "wb") as f:
                f.write(self.serialize(data))

    @staticmethod
    @abstractmethod
    def serialize(stage: Stage, /) -> bytes:
        pass



from itertools import chain
from typing import Final

from ..stage import Stage
from .file import FileSlot
from .xml import XmlSlot


__all__ = ["BinSlot"]


class BinSlot(FileSlot):
    __slots__ = ()

    @staticmethod
    def compress(data: bytes, /) -> bytes:
        buffer: bytearray = bytearray(1024)
        buffer_index: int = 958
        chunk: bytearray
        data_index: int = 0
        output: Final[bytearray] = bytearray(
            b"\x00\x00\x00\x01\x00\x00\x00\x08"
            + len(data).to_bytes(4, byteorder="big")
            + b"\x00\x00\x00\x01"
        )
        reference_indices: list[int]
        test_buffer: bytearray
        test_length: int
        test_reference_indicies: list[int]
        while data_index < len(data):
            chunk = bytearray(b"\x00")
            for bit in range(8):
                if data_index >= len(data):
                    chunk[0] >>= 8 - bit
                    output.extend(chunk)
                    return output + b"\x00" * (len(output) & 1)
                if len(data) - data_index <= 2:
                    buffer[buffer_index] = data[data_index]
                    buffer_index = buffer_index + 1 & 1023
                    chunk[0] = chunk[0] >> 1 | 128
                    chunk.append(data[data_index])
                    data_index += 1
                    continue
                reference_indices = []
                for i in chain(range(buffer_index, 1024), range(buffer_index)):
                    if data[data_index] == buffer[i]:
                        reference_indices.append(i)
                if not reference_indices:
                    buffer[buffer_index] = data[data_index]
                    buffer_index = buffer_index + 1 & 1023
                    chunk[0] = chunk[0] >> 1 | 128
                    chunk.append(data[data_index])
                    data_index += 1
                    continue
                test_buffer = buffer.copy()
                test_buffer[buffer_index] = data[data_index]
                for i in reference_indices.copy():
                    if data[data_index + 1] != test_buffer[i - 1023]:
                        reference_indices.remove(i)
                if not reference_indices:
                    buffer[buffer_index] = data[data_index]
                    buffer_index = buffer_index + 1 & 1023
                    chunk[0] = chunk[0] >> 1 | 128
                    chunk.append(data[data_index])
                    data_index += 1
                    continue
                test_buffer[buffer_index - 1023] = data[data_index + 1]
                for i in reference_indices.copy():
                    if data[data_index + 2] != test_buffer[i - 1022]:
                        reference_indices.remove(i)
                if not reference_indices:
                    buffer[buffer_index] = data[data_index]
                    buffer_index = buffer_index + 1 & 1023
                    chunk[0] = chunk[0] >> 1 | 128
                    chunk.append(data[data_index])
                    data_index += 1
                    continue
                test_length = 4
                test_reference_indicies = reference_indices.copy()
                while test_length <= min(66, len(data) - data_index):
                    test_buffer[buffer_index + test_length - 1026] = data[
                        data_index + test_length - 2
                    ]
                    for i in test_reference_indicies.copy():
                        if (
                            data[data_index + test_length - 1]
                            != test_buffer[i + test_length - 1025]
                        ):
                            test_reference_indicies.remove(i)
                    if test_reference_indicies:
                        reference_indices = test_reference_indicies.copy()
                    else:
                        break
                    test_length += 1
                chunk[0] >>= 1
                test_length -= 1
                if buffer_index + test_length >= 1024:
                    buffer[buffer_index:] = data[
                        data_index : data_index + 1024 - buffer_index
                    ]
                    buffer[: buffer_index + test_length - 1024] = data[
                        data_index + 1024 - buffer_index : data_index + test_length
                    ]
                else:
                    buffer[buffer_index : buffer_index + test_length] = data[
                        data_index : data_index + test_length
                    ]
                buffer_index = buffer_index + test_length & 1023
                chunk.extend(
                    (
                        reference_indices[0] & 255,
                        reference_indices[0] >> 2 & 192 | test_length - 3,
                    )
                )
                data_index += test_length
            output.extend(chunk)
        return bytes(output)

    @staticmethod
    def decompress(data: bytes, /) -> bytes:
        buffer: Final[bytearray] = bytearray(1024)
        buffer_index: int = 958
        handle: int | bytearray
        flags: int
        offset: int
        raw: Final[bytearray] = bytearray(data[:15:-1])
        ref: bytes
        result: Final[bytearray] = bytearray()
        result_size: Final[int] = int.from_bytes(data[8:12], byteorder="big")
        while len(result) < result_size:
            flags = raw.pop()
            for _ in range(8):
                if flags & 1:
                    handle = raw.pop()
                    buffer[buffer_index] = handle
                    buffer_index = buffer_index + 1 & 1023
                    result.append(handle)
                else:
                    if len(raw) < 2:
                        return result
                    ref = bytes((raw.pop() for _ in range(2)))
                    offset = (ref[1] << 2 & 768) + ref[0]
                    handle = bytearray()
                    for i in range((ref[1] & 63) + 3):
                        handle.append(buffer[offset + i - 1024])
                        buffer[buffer_index] = handle[-1]
                        buffer_index = buffer_index + 1 & 1023
                    result.extend(handle)
                flags >>= 1
        return bytes(result)

    @staticmethod
    def deserialize(data: bytes, /) -> Stage:
        return XmlSlot.deserialize(BinSlot.decompress(data))

    @staticmethod
    def serialize(stage: Stage, /) -> bytes:
        return BinSlot.compress(XmlSlot.serialize(stage))



from enum import Enum, unique
from typing import TypeAlias


__all__ = ["DecorationModel", "DeviceModel", "Model", "PartModel"]


@unique
class PartModel(Enum):
    Tile10x10 = 9
    Tile20x20 = 11
    TileA30x30 = 13
    TileB30x30 = 16
    Tile10x90 = 10
    Tile20x90 = 12
    TileA30x90 = 14
    TileB30x90 = 17
    Tile90x90 = 15
    MagmaTile = 20
    """Touching the magma will start you over."""
    SlipperyTile = 21
    """This tile made of ice will make you slide."""
    StickyTile = 22
    """Touching this tile makes it hard to roll."""
    Wall10x10 = 48
    Wall10x20 = 49
    Wall10x30 = 50
    Wall10x90 = 51
    Wall5x30 = 52
    Wall30x15 = 53
    DecorativeWallA = 55
    DecorativeWallB = 54
    InvisibleTile = 34
    HillA = 41
    HillB = 42
    HillC = 43
    ObstacleHill = 44
    HillWallA = 31
    HillWallB = 32
    HillWallC = 33
    Arch = 39
    ArchWall = 40
    CurveS = 1
    CurveM = 2
    CurveL = 3
    SpiralSet = 46
    RoundPillar = 4
    CurveWallS = 5
    CurveWallM = 6
    CurveWallL = 7
    SpiralWallA = 37
    SpiralWallB = 38
    StraightRailA = 60
    StraightRailB = 61
    CurveRail = 56
    ArchRail = 58
    StraightRailLA = 62
    StraightRailLB = 63
    CurveRailL = 57
    ArchRailL = 59
    FunnelPipe = 23
    StraightPipe = 24
    CurvePipe = 25
    SpiralPipe = 26
    Tunnel = 47
    Bridge = 0
    RampA = 35
    RampB = 36
    Turnover = 8
    DecorativeTile = 18
    GuardTile = 19
    MouseHole = 45
    Hole30x30 = 27
    HoleA90x90 = 28
    HoleB90x90 = 29
    HoleC90x90 = 30


@unique
class DecorationModel(Enum):
    """Decoration models are stored in a separate file; their appearance, size, shape, and availability varies depending on the theme."""

    Decoration01 = 0
    Decoration02 = 1
    Decoration03 = 2
    Decoration04 = 3
    Decoration05 = 4
    Decoration06 = 5
    Decoration07 = 6
    Decoration08 = 7
    Decoration09 = 8
    Decoration10 = 9
    Decoration11 = 10
    Decoration12 = 11


@unique
class DeviceModel(Enum):
    Start = 0
    """This part marks the starting point."""
    Goal = 1
    """This part marks the goal point."""
    Crystal = 49
    """Crystals are placed throughout the stage."""
    Respawn = 2
    """This part marks the restart point."""
    MovingTile10x10 = 3
    """A tile that moves."""
    MovingTile20x20 = 4
    """A tile that moves."""
    MovingTile30x30 = 5
    """A tile that moves."""
    MovingTile30x90 = 6
    """A tile that moves."""
    MovingTile90x90A = 8
    """A tile that moves."""
    MovingTile90x90B = 7
    """A tile that moves."""
    MovingTile10x10Switch = 84
    """This tile moves when you get on it."""
    MovingTile20x20Switch = 85
    """This tile moves when you get on it."""
    MovingTile30x30Switch = 86
    """This tile moves when you get on it."""
    MovingTile30x90Switch = 87
    """This tile moves when you get on it."""
    MovingTile90x90ASwitch = 89
    """This tile moves when you get on it."""
    MovingTile90x90BSwitch = 88
    """This tile moves when you get on it."""
    MovingFunnelPipe = 37
    """A moving pipe."""
    MovingStraightPipe = 38
    """A moving pipe."""
    MovingCurveS = 46
    """A moving curve."""
    MovingCurveM = 47
    """A moving curve."""
    MovingCurveL = 48
    """A moving curve."""
    SlidingTile = 11
    """This tile slides as you tilt the Wii Remote."""
    ConveyorBelt = 23
    """This conveyor belt carries your ball."""
    EndMagnet = 20
    """Your ball can stick to this magnet to travel."""
    StraightMagnet = 17
    """Your ball can stick to this magnet to travel."""
    CurveMagnetL = 18
    """Your ball can stick to this magnet to travel."""
    CurveMagnetS = 19
    """Your ball can stick to this magnet to travel."""
    DashTunnelA = 32
    """This tunnel speeds up your ball."""
    DashTunnelB = 33
    """This tunnel speeds up your ball."""
    SeesawLBlock = 13
    """This part rotates when tilting the Wii Remote."""
    SeesawIBlock = 14
    """This part rotates when tilting the Wii Remote."""
    AutoSeesawLBlock = 15
    """A rotating block."""
    AutoSeesawIBlock = 16
    """A rotating block."""
    Cannon = 21
    """Fires your ball into the air."""
    Drawbridge = 44
    """Can be crossed if your ball bumps it down."""
    Turntable = 35
    """A rotating disc."""
    Bumper = 12
    """Your ball goes flying when touching it."""
    PowerfulBumper = 51
    """Your ball goes flying when touching it."""
    Thorn = 34
    """You have to start over when touching it."""
    Gear = 39
    """A rotating gear."""
    Fan = 45
    """This fan will blow your ball around."""
    PowerfulFan = 52
    """A fan with strong wind power."""
    TimerFan = 53
    """Its wind power is switched on and off."""
    Spring = 30
    """This tile makes your ball bounce."""
    Punch = 27
    """Pushes the ball forward with a strong punch."""
    Press = 28
    """Pushes the ball forward."""
    Scissors = 36
    """You'll start over if you ball is chopped."""
    MagnifyingGlass = 29
    """You start over if the ball hits the light."""
    UpsideDownStageDevice = 31
    """Flips the stage 180 degrees upside down."""
    UpsideDownBall = 43
    """Flips gravity upside down."""
    SmallTunnel = 25
    """Your ball shrinks when going through."""
    BigTunnel = 26
    """Your ball grows when going through."""
    ToyTrain = 58
    """Carries your ball."""
    EndTracks = 57
    LeftTracks = 56
    RightTracks = 55
    StraightTracks = 54
    Warp = 41
    """Teleports your ball to another warp point."""
    BlinkingTile = 40
    """Turns visible and invisible."""
    MelodyTileLowG = 59
    """Makes a sound when the ball rolls over it."""
    MelodyTileLowGSharp = 60
    """Makes a sound when the ball rolls over it."""
    MelodyTileLowA = 61
    """Makes a sound when the ball rolls over it."""
    MelodyTileLowASharp = 62
    """Makes a sound when the ball rolls over it."""
    MelodyTileLowB = 63
    """Makes a sound when the ball rolls over it."""
    MelodyTileC = 64
    """Makes a sound when the ball rolls over it."""
    MelodyTileCSharp = 65
    """Makes a sound when the ball rolls over it."""
    MelodyTileD = 66
    """Makes a sound when the ball rolls over it."""
    MelodyTileDSharp = 67
    """Makes a sound when the ball rolls over it."""
    MelodyTileE = 68
    """Makes a sound when the ball rolls over it."""
    MelodyTileF = 69
    """Makes a sound when the ball rolls over it."""
    MelodyTileFSharp = 70
    """Makes a sound when the ball rolls over it."""
    MelodyTileG = 71
    """Makes a sound when the ball rolls over it."""
    MelodyTileGSharp = 72
    """Makes a sound when the ball rolls over it."""
    MelodyTileA = 73
    """Makes a sound when the ball rolls over it."""
    MelodyTileASharp = 74
    """Makes a sound when the ball rolls over it."""
    MelodyTileB = 75
    """Makes a sound when the ball rolls over it."""
    MelodyTileHighC = 76
    """Makes a sound when the ball rolls over it."""
    MelodyTileHighCSharp = 77
    """Makes a sound when the ball rolls over it."""
    MelodyTileHighD = 78
    """Makes a sound when the ball rolls over it."""
    MelodyTileHighDSharp = 79
    """Makes a sound when the ball rolls over it."""
    MelodyTileHighE = 80
    """Makes a sound when the ball rolls over it."""
    MelodyTileHighF = 81
    """Makes a sound when the ball rolls over it."""
    MelodyTileHighFSharp = 82
    """Makes a sound when the ball rolls over it."""
    MelodyTileHighG = 83
    """Makes a sound when the ball rolls over it."""
    KororinCapsule = 95
    GreenCrystal = 96
    Ant = 97

    # Extra models
    MovingTile20x20Wall = 9
    """The wall that can be added to Moving Tile 20x20"""
    MovingTile30x30Wall = 10
    """The wall that can be added to Moving Tile 30x30"""
    CannonArrow = 22
    """The arrow used to show the trajectory from a Cannon"""
    ConveyorBeltArrows = 24
    """The arrows used by the Conveyor Belt when changing direction"""
    BlueLine = 42
    """The blue line used by moving tiles"""
    OrangeLine = 50
    """The orange line used by switch moving tiles"""
    CubicTextBox = 92
    """A cubic region that opens a text box"""
    WallTextBox = 93
    """A tall, wide, and thin region that opens a text box"""
    QuestionMark = 94
    """An animated question mark with a glow beneath it"""


Model: TypeAlias = PartModel | DecorationModel | DeviceModel
